'''
Implement CopulaGAN model for tabular simulation
prediction in clinical trials.
'''
import os
import pdb
import warnings
import joblib

import torch
from torch import optim
import numpy as np
import pandas as pd
from ctgan import CTGANSynthesizer as CTGANSynthesizerBase
from ctgan.data_sampler import DataSampler
from ctgan.data_transformer import DataTransformer as DataTransformerBase
from ctgan.synthesizers.ctgan import Generator, Discriminator

from .base import TabularSimulationBase
from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist

warnings.filterwarnings('ignore')

class DataTransformer(DataTransformerBase):
    def transform(self, raw_data):
        """Take raw data and output a matrix data."""
        if not isinstance(raw_data, pd.DataFrame):
            column_names = [str(num) for num in range(raw_data.shape[1])]
            raw_data = pd.DataFrame(raw_data, columns=column_names)

        # Only use parallelization with larger data sizes.
        # Otherwise, the transformation will be slower.
        # if raw_data.shape[0] < 500:
        #     column_data_list = self._synchronous_transform(
        #         raw_data,
        #         self._column_transform_info_list
        #     )
        # else:
        #     column_data_list = self._parallel_transform(
        #         raw_data,
        #         self._column_transform_info_list
        #     )


        # do not use parallelization cuz bugs
        column_data_list = self._synchronous_transform(
            raw_data,
            self._column_transform_info_list
        )

        return np.concatenate(column_data_list, axis=1).astype(float)


class CTGANSynthesizer(CTGANSynthesizerBase):
    def fit(self, train_data, discrete_columns=(), epochs=None):
        """Fit the CTGAN Synthesizer models to the training data.

        Args:
            train_data (numpy.ndarray or pandas.DataFrame):
                Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
            discrete_columns (list-like):
                List of discrete columns to be used to generate the Conditional
                Vector. If ``train_data`` is a Numpy array, this list should
                contain the integer indices of the columns. Otherwise, if it is
                a ``pandas.DataFrame``, this list should contain the column names.
        """
        self._validate_discrete_columns(train_data, discrete_columns)
        if epochs is None:
            epochs = self._epochs
        else:
            warnings.warn(
                ('`epochs` argument in `fit` method has been deprecated and will be removed '
                 'in a future version. Please pass `epochs` to the constructor instead'),
                DeprecationWarning
            )

        # skip transformation beacuse we have already transformed the data
        self._transformer = DataTransformer()
        self._transformer.fit(train_data, discrete_columns)
        train_data = self._transformer.transform(train_data)

        self._data_sampler = DataSampler(
            train_data,
            self._transformer.output_info_list,
            self._log_frequency)

        data_dim = self._transformer.output_dimensions

        self._generator = Generator(
            self._embedding_dim + self._data_sampler.dim_cond_vec(),
            self._generator_dim,
            data_dim
        ).to(self._device)

        discriminator = Discriminator(
            data_dim + self._data_sampler.dim_cond_vec(),
            self._discriminator_dim,
            pac=self.pac
        ).to(self._device)

        optimizerG = optim.Adam(
            self._generator.parameters(), lr=self._generator_lr, betas=(0.5, 0.9),
            weight_decay=self._generator_decay
        )

        optimizerD = optim.Adam(
            discriminator.parameters(), lr=self._discriminator_lr,
            betas=(0.5, 0.9), weight_decay=self._discriminator_decay
        )

        mean = torch.zeros(self._batch_size, self._embedding_dim, device=self._device)
        std = mean + 1

        steps_per_epoch = max(len(train_data) // self._batch_size, 1)
        for i in range(epochs):
            for id_ in range(steps_per_epoch):

                for n in range(self._discriminator_steps):
                    fakez = torch.normal(mean=mean, std=std)

                    condvec = self._data_sampler.sample_condvec(self._batch_size)
                    if condvec is None:
                        c1, m1, col, opt = None, None, None, None
                        real = self._data_sampler.sample_data(self._batch_size, col, opt)
                    else:
                        c1, m1, col, opt = condvec
                        c1 = torch.from_numpy(c1).to(self._device)
                        m1 = torch.from_numpy(m1).to(self._device)
                        fakez = torch.cat([fakez, c1], dim=1)

                        perm = np.arange(self._batch_size)
                        np.random.shuffle(perm)
                        real = self._data_sampler.sample_data(
                            self._batch_size, col[perm], opt[perm])
                        c2 = c1[perm]

                    fake = self._generator(fakez)
                    fakeact = self._apply_activate(fake)

                    real = torch.from_numpy(real.astype('float32')).to(self._device)

                    if c1 is not None:
                        fake_cat = torch.cat([fakeact, c1], dim=1)
                        real_cat = torch.cat([real, c2], dim=1)
                    else:
                        real_cat = real
                        fake_cat = fakeact

                    y_fake = discriminator(fake_cat)
                    y_real = discriminator(real_cat)

                    pen = discriminator.calc_gradient_penalty(
                        real_cat, fake_cat, self._device, self.pac)
                    loss_d = -(torch.mean(y_real) - torch.mean(y_fake))

                    optimizerD.zero_grad()
                    pen.backward(retain_graph=True)
                    loss_d.backward()
                    optimizerD.step()

                fakez = torch.normal(mean=mean, std=std)
                condvec = self._data_sampler.sample_condvec(self._batch_size)

                if condvec is None:
                    c1, m1, col, opt = None, None, None, None
                else:
                    c1, m1, col, opt = condvec
                    c1 = torch.from_numpy(c1).to(self._device)
                    m1 = torch.from_numpy(m1).to(self._device)
                    fakez = torch.cat([fakez, c1], dim=1)

                fake = self._generator(fakez)
                fakeact = self._apply_activate(fake)

                if c1 is not None:
                    y_fake = discriminator(torch.cat([fakeact, c1], dim=1))
                else:
                    y_fake = discriminator(fakeact)

                if condvec is None:
                    cross_entropy = 0
                else:
                    cross_entropy = self._cond_loss(fake, c1, m1)

                loss_g = -torch.mean(y_fake) + cross_entropy

                optimizerG.zero_grad()
                loss_g.backward()
                optimizerG.step()

            if self._verbose:
                print(f'Epoch {i+1}, Loss G: {loss_g.detach().cpu(): .4f},'  # noqa: T001
                      f'Loss D: {loss_d.detach().cpu(): .4f}',
                      flush=True)


class BuildModel:
    def __new__(self, config) -> CTGANSynthesizer:
        model = CTGANSynthesizer(
            embedding_dim=config['embedding_dim'],
            generator_dim=config['generator_dim'],
            discriminator_dim=config['discriminator_dim'],
            generator_lr=config['generator_lr'],
            generator_decay=config['generator_decay'],
            discriminator_lr=config['discriminator_lr'],
            discriminator_decay=config['discriminator_decay'],
            batch_size=config['batch_size'],
            discriminator_steps=config['discriminator_steps'],
            log_frequency=config['log_frequency'],
            verbose=config['verbose'],
            epochs=config['epochs'],
            pac=config['pac'],
            cuda=config['cuda'],
            )
        return model


class CopulaGAN(TabularSimulationBase):
    '''
    Implement CopulaGAN model for tabular patient data simulation [1]_.

    Parameters
    ----------
    embedding_dim: int
        Size of the random sample passed to the Generator. Defaults to 128.

    generator_dim: tuple or list of int:
        Size of the output samples for each one of the Residuals. A Residual Layer
        will be created for each one of the values provided. Defaults to (256, 256).
    
    discriminator_dim: tuple or list of ints
        Size of the output samples for each one of the Discriminator Layers. A Linear Layer
        will be created for each one of the values provided. Defaults to (256, 256).
    
    generator_lr: float
        Learning rate for the generator. Defaults to 2e-4.
    
    generator_decay: float
        Generator weight decay for the Adam Optimizer. Defaults to 1e-6.
    
    discriminator_lr: float
        Learning rate for the discriminator. Defaults to 2e-4.
    
    discriminator_decay: float
        Discriminator weight decay for the Adam Optimizer. Defaults to 1e-6.
    
    batch_size: int
        Number of data samples to process in each step.
    
    discriminator_steps: int
        Number of discriminator updates to do for each generator update.
        From the WGAN paper: https://arxiv.org/abs/1701.07875. WGAN paper
        default is 5. Default used is 1 to match original CTGAN implementation.
    
    log_frequency: boolean
        Whether to use log frequency of categorical levels in conditional
        sampling. Defaults to ``True``.
    
    verbose: boolean
        Whether to have print statements for progress results. Defaults to ``True``.
    
    epochs: int
        Number of training epochs. Defaults to 10.
    
    pac: int
        Number of samples to group together when applying the discriminator.
        Defaults to 10.
    
    cuda: bool or str
        - If ``True``, use CUDA. If a ``str``, use the indicated device.
        - If ``False``, do not use cuda at all.
        
    experiment_id: str, optional
        The name of current experiment. Decide the saved model checkpoint name.

    Notes
    -----
    .. [1] Xu, L., Skoularidou, M., Cuesta-Infante, A., & Veeramachaneni, K. (2019). Modeling tabular data using conditional gan. Advances in Neural Information Processing Systems, 32.
    '''
    def __init__(
        self,
        embedding_dim=128,
        generator_dim=(256, 256),
        discriminator_dim=(256, 256),
        generator_lr=2e-4,
        generator_decay=1e-6,
        discriminator_lr=2e-4,
        discriminator_decay=1e-6,
        batch_size=500,
        discriminator_steps=1,
        log_frequency=True,
        verbose=True,
        epochs=50,
        pac=10,
        cuda=False, # can be set to "True" if applicable
        experiment_id='trial_simulation.tabular.copulagan',
    ) -> None:
        super().__init__(experiment_id=experiment_id)
        self.config = {
            'embedding_dim' : embedding_dim,
            'generator_dim' : generator_dim,
            'discriminator_dim' : discriminator_dim,
            'generator_lr' : generator_lr,
            'generator_decay' : generator_decay,
            'discriminator_lr' : discriminator_lr,
            'discriminator_decay' : discriminator_decay,
            'batch_size' : batch_size,
            'discriminator_steps' : discriminator_steps,
            'log_frequency' : log_frequency,
            'verbose' : verbose,
            'epochs' : epochs,
            'pac' : pac,
            'cuda' : cuda,
            'experiment_id': experiment_id,
            'model_name': 'copula_gan',
        }
        self._save_config(self.config)
        self._build_model()

    def fit(self, train_data):
        '''
        Train CopulaGAN model to simulate patient outcome
        with tabular input data.

        Parameters
        ----------
        train_data: TabularPatientBase
            The training data for the model.
        '''
        self._input_data_check(train_data)
        if isinstance(train_data, TabularPatientBase): # transform=True
            dataset = train_data.df
        if isinstance(train_data, dict): 
            dataset = TabularPatientBase(train_data, transform=True)
            dataset = dataset.df
        self._fit_model(dataset)
        self.metadata = train_data.metadata
        self.raw_dataset = train_data
        
    def predict(self, n=200):
        '''
        Simulate new tabular data with number_of_predictions.

        Parameters
        ----------
        n: int
            The number of synthetic data to generate.

        Returns
        -------
        ypred: TabularPatientBase
            A new tabular data simulated by the model
        '''
        ypred = self.model.sample(n)  # build df
        ypred = self.raw_dataset.reverse_transform(ypred) # transform back
        return ypred # output: dataset, same as the input dataset

    def save_model(self, output_dir=None):
        '''
        Save the learned CopulaGAN model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        self._save_config(self.config, output_dir=output_dir)
        ckpt_path = os.path.join(output_dir, 'copulagan.model')
        joblib.dump(self.model, ckpt_path)

    def load_model(self, checkpoint=None):
        '''
        Save the learned CopulaGAN model to the disk.

        Parameters
        ----------
        checkpoint: str or None
            The path to the saved model.
            - If a directory, the only checkpoint file `.model` will be loaded.
            - If a filepath, will load from this file;
            - If None, will load from `self.checkout_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkout_dir

        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        config_filename = check_model_config_file(checkpoint)
        self.model = joblib.load(checkpoint_filename)
        self.config = self._load_config(config_filename)

    def _build_model(self):
        self.model = BuildModel(self.config)

    def _fit_model(self, data):
        self.model.fit(data)



import torch
import torch.nn as nn
import torch.nn.functional as F
from copy import deepcopy 
from torch.autograd import Variable
from torch.utils import data
from torch.utils.data import SequentialSampler
import matplotlib.pyplot as plt
import numpy as np 
sigmoid = torch.nn.Sigmoid() 
torch.manual_seed(0)

from .gnn_layers import GraphConvolution, GraphAttention
torch.manual_seed(4) 
np.random.seed(1)

class Highway(nn.Module):
    def __init__(self, size, num_layers):
        super(Highway, self).__init__()
        self.num_layers = num_layers
        self.nonlinear = nn.ModuleList([nn.Linear(size, size) for _ in range(num_layers)])
        self.linear = nn.ModuleList([nn.Linear(size, size) for _ in range(num_layers)])
        self.gate = nn.ModuleList([nn.Linear(size, size) for _ in range(num_layers)])
        self.f = F.relu

    def forward(self, x):
        """
            :param x: tensor with shape of [batch_size, size]
            :return: tensor with shape of [batch_size, size]
            applies σ(x) ⨀ (f(G(x))) + (1 - σ(x)) ⨀ (Q(x)) transformation | G and Q is affine transformation,
            f is non-linear transformation, σ(x) is affine transformation with sigmoid non-linearition
            and ⨀ is element-wise multiplication
        """
        for layer in range(self.num_layers):
            gate = F.sigmoid(self.gate[layer](x))
            nonlinear = self.f(self.nonlinear[layer](x))
            linear = self.linear[layer](x)
            x = gate * nonlinear + (1 - gate) * linear
        return x






class GCN(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, init):
        super(GCN, self).__init__()

        self.gc1 = GraphConvolution(nfeat, nhid, init=init)
        self.gc2 = GraphConvolution(nhid, nclass, init=init)
        self.dropout = dropout

    def bottleneck(self, path1, path2, path3, adj, in_x):
        return F.relu(path3(F.relu(path2(F.relu(path1(in_x, adj)), adj)), adj))

    def forward(self, x, adj):
        x = F.dropout(F.relu(self.gc1(x, adj)), self.dropout, training=self.training)
        x = self.gc2(x, adj)
        return x 
        # return F.log_softmax(x, dim=1)




class GCN_drop_in(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, init):
        super(GCN_drop_in, self).__init__()

        self.gc1 = GraphConvolution(nfeat, nhid, init=init)
        self.gc2 = GraphConvolution(nhid, nclass, init=init)
        self.dropout = dropout

    def bottleneck(self, path1, path2, path3, adj, in_x):
        return F.relu(path3(F.relu(path2(F.relu(path1(in_x, adj)), adj)), adj))

    def forward(self, x, adj):
        x = F.dropout(x, self.dropout, training=self.training)
        x = F.dropout(F.relu(self.gc1(x, adj)), self.dropout, training=self.training)
        x = self.gc2(x, adj)

        return F.log_softmax(x, dim=1)

class GAT(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, alpha, nheads):
        super(GAT, self).__init__()
        self.dropout = dropout

        self.attentions = [GraphAttention(nfeat, nhid, dropout=dropout, alpha=alpha, concat=True) for _ in range(nheads)]
        for i, attention in enumerate(self.attentions):
            self.add_module('attention_{}'.format(i), attention)

        self.out_att = GraphAttention(nhid * nheads, nclass, dropout=dropout, alpha=alpha, concat=False)

    def forward(self, x, adj):
        x = F.dropout(x, self.dropout, training=self.training)
        x = torch.cat([att(x, adj) for att in self.attentions], dim=1)
        x = F.dropout(x, self.dropout, training=self.training)
        x = F.elu(self.out_att(x, adj))
        return F.log_softmax(x, dim=1)




if __name__ == "__main__":
    gnn = GCN(
            nfeat = 20,
            nhid = 30,
            nclass = 1,
            dropout = 0.6,
            init = 'uniform') 












import pdb

import torch
from torch import nn

class COMPOSELoss(nn.Module):
    def __init__(self, model) -> None:
        super().__init__()
        self.ce_loss = nn.CrossEntropyLoss(reduction='none')
        self.sm_loss = nn.CosineEmbeddingLoss(margin=0.3, reduction='none')
        self.model = model

    def forward(self, inputs):
        # forward
        pred = self.model(inputs)

        # compute label loss
        inc_ce_loss = self._compute_label_loss(pred, inputs, 'inc')
        exc_ce_loss = self._compute_label_loss(pred, inputs, 'exc')

        # compute embedding loss
        sm_loss_inc = self._compute_embed_loss(pred, inputs, 'inc')
        sm_loss_exc = self._compute_embed_loss(pred, inputs, 'exc')

        # all loss
        loss_value = (inc_ce_loss + exc_ce_loss + sm_loss_inc + sm_loss_exc) / 2

        return {'loss_value':loss_value}

    def _compute_label_loss(self, pred, inputs, ec='inc'):
        pred_inc = pred[f'logit_{ec}'] # bs_trial, bs_ehr, num_ec, num_label
        num_sample = min(pred_inc.shape[0], pred_inc.shape[1])        
        pred_inc_logit = pred_inc[range(num_sample),range(num_sample)].permute(0,2,1) # bs_trial==bs_patient, num_label, num_ec
        inc_label = inputs[f'{ec}_label'] # bs_trial, num_ec
        inc_label_mask = inputs[f'{ec}_emb_mask']
        
        # mask out unknown predictions to prevent model from making unknown predictions all the time
        unk_label_mask = torch.ones(inc_label_mask.shape, device=inc_label_mask.device)
        unk_label_mask[inc_label == 2] = 0
        inc_label_mask = inc_label_mask * unk_label_mask

        # compute masked loss
        inc_ce_loss = self.ce_loss(pred_inc_logit, inc_label.long())
        inc_ce_loss = torch.sum(inc_ce_loss * inc_label_mask) / inc_label_mask.sum()
        return inc_ce_loss

    def _compute_embed_loss(self, pred, inputs, ec='inc'):
        response_inc = pred[f'response_{ec}']
        query_inc = pred[f'query_{ec}']
        num_sample = min(response_inc.shape[0], response_inc.shape[1])
        response_inc_emb = response_inc[range(num_sample), range(num_sample)] # bs, num_ec, emb_dim
        query_inc_emb = query_inc[range(num_sample), range(num_sample)] # bs, num_ec, emb_dim
        inc_label = inputs[f'{ec}_label'] # bs_trial, num_ec
        inc_label_mask = inputs[f'{ec}_emb_mask']

        # build similarity loss label
        similarity_label = torch.ones(inc_label.shape, device=inc_label.device)
        unk_label_mask = torch.ones(inc_label.shape, device=inc_label.device)
        similarity_label[inc_label == 0] = -1 # unmatched
        similarity_label[inc_label == 1] = 1 # matched
        similarity_label[inc_label == 2] = 1
        unk_label_mask[inc_label == 2] = 0
        inc_label_mask = inc_label_mask * unk_label_mask

        sm_loss = self.sm_loss(response_inc_emb.view(-1, response_inc_emb.shape[-1]), query_inc_emb.view(-1, query_inc_emb.shape[-1]), similarity_label.view(-1,))
        sm_loss = torch.sum(inc_label_mask.view(-1) * sm_loss) / inc_label_mask.sum()

        return sm_loss


class DeepEnrollLoss(nn.Module):
    def __init__(self, model) -> None:
        super().__init__()
        self.ce_loss = nn.CrossEntropyLoss(reduction='none')
        self.model = model

    def forward(self, inputs):
        # forward
        pred = self.model(inputs)

        # compute label loss
        inc_ce_loss = self._compute_label_loss(pred, inputs, 'inc')
        exc_ce_loss = self._compute_label_loss(pred, inputs, 'exc')

        # all loss
        loss_value = (inc_ce_loss + exc_ce_loss) / 2

        return {'loss_value':loss_value}

    def _compute_label_loss(self, pred, inputs, ec='inc'):
        pred_inc = pred[f'logit_{ec}'] # bs_trial, bs_ehr, num_ec, num_label
        num_sample = min(pred_inc.shape[0], pred_inc.shape[1])        
        pred_inc_logit = pred_inc[range(num_sample),range(num_sample)].permute(0,2,1) # bs_trial==bs_patient, num_label, num_ec
        inc_label = inputs[f'{ec}_label'] # bs_trial, num_ec
        inc_label_mask = inputs[f'{ec}_emb_mask']
        
        # mask out unknown predictions to prevent model from making unknown predictions all the time
        unk_label_mask = torch.ones(inc_label_mask.shape, device=inc_label_mask.device)
        unk_label_mask[inc_label == 2] = 0
        inc_label_mask = inc_label_mask * unk_label_mask

        # compute masked loss
        inc_ce_loss = self.ce_loss(pred_inc_logit, inc_label.long())
        inc_ce_loss = torch.sum(inc_ce_loss * inc_label_mask) / inc_label_mask.sum()
        return inc_ce_loss




import abc
import os
import json
import pdb

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset, DataLoader

from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import check_model_dir
from pytrial.utils.check import make_dir_if_not_exist
from pytrial.utils.parallel import batch_to_device

from ..losses import BinaryXentLoss, XentLoss, MSELoss, MultilabelBinaryXentLoss
from ..trainer import IndivTabTrainer


class TabularIndivBase(abc.ABC):
    '''Abstract class for all individual outcome predictions
    based on tabular patient data.

    Parameters
    ----------
    experiment_id: str, optional (default = 'test')
        The name of current experiment.
    '''
    @abc.abstractmethod
    def __init__(self, experiment_id='test'):
        check_model_dir(experiment_id)
        self.checkout_dir = os.path.join('./experiments_records', experiment_id,
                                         'checkpoints')
        self.result_dir = os.path.join('./experiments_records', experiment_id,
                                       'results')
        make_dir_if_not_exist(self.checkout_dir)
        make_dir_if_not_exist(self.result_dir)

    @abc.abstractmethod
    def fit(self, train_data, valid_data):
        '''
        Fit function needs to be implemented after subclass.

        Parameters
        ----------
        train_data: Any
            Training data.
        
        valid_data: Any
            Validation data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, test_data):
        '''
        Prediction function needs to be implemented after subclass.

        Parameters
        ----------
        test_data: Any
            Testing data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, checkpoint):
        '''
        Load the pretrained model from disk, needs to be implemented after subclass.
        
        Parameters
        ----------
        checkpoint: str
            The path to the checkpoint file.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, output_dir):
        '''
        Save the model to disk, needs to be implemented after subclass.

        Parameters
        ----------
        output_dir: str
            The path to the output directory.
        '''
        raise NotImplementedError

    def train(self, mode=True):
        '''Set the model in training mode. Work similar to `model.train()` in PyTorch.
        
        Parameters
        ----------
        mode: bool, optional (default = True)
            Whether to set the model in training mode.
            ``False`` means the model is in evaluation mode.
            ``True`` means the model is in training mode.
        '''
        self.training = mode
        self.model.train()
        return self
    
    def eval(self, mode=False):
        '''Set the model in evaluation mode. Work similar to `model.eval()` in PyTorch.

        Parameters
        ----------
        mode: bool, optional (default = False)
            Whether to set the model in evaluation mode.
            ``False`` means the model is in evaluation mode.
            ``True`` means the model is in training mode.
        '''
        self.training = mode
        self.model.eval()
        return self

    def get_train_dataloader(self, inputs):
        dataset = self._build_dataset(inputs)
        dataloader = DataLoader(dataset, 
            batch_size=self.config['batch_size'], 
            shuffle=True, 
            num_workers=self.config['num_worker'],
            pin_memory=True,
            )
        return dataloader

    def _build_dataset(self, inputs):
        df = inputs['x']
        
        if 'y' in inputs: label = inputs['y']
        else: label = None

        return IndivTabDataset(df, label)

    def _build_loss_model(self):
        mode = self.config['mode']
        if mode == 'multiclass':
            return XentLoss(self.model)

        elif mode == 'binary':
            return BinaryXentLoss(self.model)

        elif mode == 'regression':
            return MSELoss(self.model)

        elif mode == 'multilabe':
            return MultilabelBinaryXentLoss(self.model)

        else:
            raise ValueError(f'Do not recognize mode `{mode}`, please correct.')

    def _fit_model(self, train_data, valid_data=None):
        train_dataloader = self.get_train_dataloader(train_data)
        loss_model = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model)]

        mode = self.config['mode']
        test_metric_dict = {
            'binary': 'auc',
            'multiclass': 'acc',
            'regression': 'mse',
            'multilabel': 'f1', # take average of F1 scores
        }

        if mode == 'regression':
            less_is_better = True
        else:
            less_is_better = False

        trainer = IndivTabTrainer(
            model=self,
            train_objectives=train_objectives,
            test_data=valid_data,
            test_metric=test_metric_dict[mode],
            less_is_better=less_is_better,
        )

        trainer.train(
            **self.config
        )

    def _parse_input_data(self, inputs):
        if isinstance(inputs, dict):
            if isinstance(inputs['x'], TabularPatientBase):
                dataset = inputs['x']
                x_feat = dataset.df
                y = inputs['y']
            else:
                x_feat = inputs['x']
                y = inputs['y']                
            
            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.values
        
        if isinstance(inputs, pd.DataFrame) or isinstance(inputs, torch.Tensor):
            x_feat, y = inputs, None
        
        if isinstance(inputs, TabularPatientBase):
            x_feat, y = inputs.df, None

        return x_feat, y

    def _predict_on_dataloader(self, dataloader):
        pred_list, label_list = [], []
        for batch in dataloader:
            x_feat = batch['x']
            x_feat = x_feat.to(self.device)
            if 'y' in batch: label_list.append(batch.pop('y'))
            pred = self.model(x_feat)
            pred_list.append(pred)
        pred = torch.cat(pred_list)
        label = torch.cat(label_list) if len(label_list) > 0 else None
        return {'pred':pred, 'label':label}

    def _save_config(self, config, output_dir=None):
        if output_dir is None:
            output_dir = self.checkout_dir

        temp_path = os.path.join(output_dir, 'config.json')
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

    def _load_config(self, checkpoint=None):
        '''
        Load model config from the given directory.

        Parameters
        ----------
        checkpoint: str
            The given filepath (e.g., ./checkpoint/config.json)
            to load the model config.
        '''
        if checkpoint is None:
            temp_path = os.path.join(checkpoint, 'config.json')
        else:
            temp_path = checkpoint

        assert os.path.exists(temp_path), 'Cannot find `config.json` under {}'.format(self.checkout_dir)
        with open(temp_path, 'r') as f:
            config = json.load(f)
        return config

    def _save_checkpoint(self, state,
                        epoch_id=0,
                        is_best=False,
                        output_dir=None,
                        filename='checkpoint.pth.tar'):
        if output_dir is None:
            output_dir = self.checkout_dir

        if epoch_id < 1:
            filepath = os.path.join(output_dir, 'latest.' + filename)
        elif is_best:
            filepath = os.path.join(output_dir, 'best.' + filename)
        else:
            filepath = os.path.join(self.checkout_dir,
                                    str(epoch_id) + '.' + filename)
        torch.save(state, filepath)

    def _input_data_check(self, inputs):
        '''
        Check the training / testing data fits the formats.
        Target to (1) check if inputs valid,
                    if not, give tips about the data problem.

        Parameters
        ----------
        inputs: {
                'x': TabularPatientBase or pd.DataFrame,
                'y': pd.Series or np.ndarray
                }
                'x' contain all patient features; 'y' contain labels
                for each row.
        '''
        if isinstance(inputs, dict):
            assert 'x' in inputs, 'No input patient data found in inputs.'
            assert isinstance(inputs['x'], pd.DataFrame) or isinstance(inputs['x'], TabularPatientBase), 'Get unaccepted input data format, expect `pd.DataFrame` or `TabularPatientBase`, get {} instead.'.format(type(inputs['x']))
            if 'y' in inputs:
                assert isinstance(inputs['y'], pd.Series) or isinstance(inputs['y'], np.ndarray)
                assert not pd.isnull(inputs['y']).any(), 'Find NaN in input targets, please check.'
        if isinstance(inputs['x'], pd.DataFrame):
            assert not inputs['x'].isnull().values.any(), 'Find NaN in input dataframe, please check your input, or try to pass `TabularPatientBase` as inputs.'
        if isinstance(inputs['x'], TabularPatientBase):
            assert not inputs['x'].df.isnull().values.any(), 'Find NaN in input dataset, please check your input, or try to pass `TabularPatientBase` as inputs.'


class IndivTabDataset(Dataset):
    def __init__(self, df, label=None):
        self.df = df
        self.label = label
    
    def __len__(self):
        return len(self.df)
    
    def __getitem__(self, idx):
        if self.label is not None: return {'x':self.df.iloc[idx].values, 'y': self.label[idx]}
        else: return {'x':self.df.iloc[idx].values}



'''
Evaluation methods for the simulated sequence of patient records.
'''
import pdb
from copy import deepcopy

from torch.utils.data import DataLoader
from sklearn.metrics import roc_auc_score
import numpy as np

from pytrial.tasks.indiv_outcome.sequence import RNN
from pytrial.data.patient_data import SeqPatientCollator
from pytrial.data.utils import concat_dataset, split_dataset
from pytrial.tasks.indiv_outcome.trainer import IndivSeqTrainer

from ..data import SequencePatient
from .base import transform_sequence_to_table
from .base import transform_table_to_sequence


__all__ = ['RNNPrivacyDetection', 'RNNUtilityDetection', 'DimWiseFidelity', 'PresenceDisclosurePrivacy']


class SequenceMetric:
    '''
    Base class for evaluation of the simulated sequence of patient records.
    '''
    @staticmethod
    def _build_dataset(real_data, syn_data):
        raise NotImplementedError

    @staticmethod
    def _build_model(train_data, test_data):
        '''
        Build a detection model.
        '''
        raise NotImplementedError
    
    @classmethod
    def compute(self, real_data, syn_data):
        raise NotImplementedError

class SequencePrivacyMetric(SequenceMetric):
    '''
    Base class for Machine Learning detection based metrics for sequential synthetic patient records.

    These metrics build a Machine Learning Classifier that learns to tell the synthetic
    data apart from the real data.

    The output is the binary classification AUC.
    '''
    @staticmethod
    def _build_dataset(real_data, syn_data):
        '''
        Build a dataset for training a ML model.

        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        Returns
        -------
        train_set: torch.utils.data.Dataset
            A dataset for training a ML model.
        
        test_set: torch.utils.data.Dataset
            A dataset for testing a ML model.
        '''
        real_syn_data = concat_dataset(real_data, syn_data)
        train_data, test_data = split_dataset(real_syn_data, split=0.8)
        return train_data, test_data
    
    @staticmethod
    def _build_model(train_data, test_data):
        '''
        Build a detection model.
        '''      
        raise NotImplementedError


class RNNPrivacyDetection(SequencePrivacyMetric):
    '''
    Try to discriminate real and synthetic patient records using an RNN based neural network.
    The output score is AUC evaluated on the prediction.
    So the biggest score will be AUC=1.0, which means the model can correctly discriminate real and synthetic data.
    The lowest score will be AUC=0.5, which means the model cannot discriminate real and sytnethic data.

    There are multiple interpretations of the score. 
    A low score (low AUC) can indicates high synthetic data quality as well as low privacy. 
    A high score (high AUC) can indicate low synthetic data quality as well as high privacy. 
    '''

    @classmethod
    def compute(self, real_data, syn_data, device='cpu'):
        '''
        Compute the detection score.

        Parameters
        ----------
        real_data: pytrial.data.patient_sequence.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.data.patient_sequence.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        device: str
            The device to run the model on. Default is 'cpu'.

        Returns
        -------
        float: The detection score.
        '''
        assert isinstance(real_data, SequencePatient), '`real_data` must be a `SequencePatient` object.'
        assert isinstance(syn_data, SequencePatient), '`syn_data` must be a `SequencePatient` object.'

        # assign labels to real and syn data
        real_data = deepcopy(real_data)
        syn_data = deepcopy(syn_data)
        real_data.label = [1] * len(real_data) # 1 is real data label
        syn_data.label = [0] * len(syn_data) # 0 is syn data label

        # build a dataset with labeled real and syndata
        train_data, test_data = self._build_dataset(real_data, syn_data)

        # train a RNN model
        model = self._build_model(train_data, test_data, device=device)

        # evaluate the model
        pred = model.predict(test_data)
        label = test_data.get_label()
        auc = roc_auc_score(label, pred)
        return auc

    @staticmethod
    def _build_model(train_data, test_data, device='cpu'):
        '''
        Build a RNN model for binary classification.

        Parameters
        ----------
        dataloader: torch.utils.data.DataLoader
            A dataloader for training a ML model.

        Returns
        -------
        A ML model for binary classification of synthetic and real patient records.
        '''
        model = RNN(
            vocab_size=[len(train_data.metadata['voc'][o]) for o in train_data.metadata['visit']['order']],
            orders=train_data.metadata['visit']['order'],
            mode='binary',
            bidirectional=False,
            epochs=10,
            batch_size=64,
            device=device,
            )
        model.fit(train_data=train_data, valid_data=test_data)
        return model


class PresenceDisclosurePrivacy(SequencePrivacyMetric):
    '''
    Try to evaluate the presence disclosure risk of the synthetic patient records.
    '''
    @classmethod
    def compute(self, real_data, syn_data, known_patient=[1,5], confidence=1, prior=1, device=None):
        '''
        Compute the presence disclosure risk. Assume the attacker knows the real patient records with `known_patient` percentage.
        The attacker will try to match the synthetic patient records with the known real patient records to check
        if the synthetic patient records disclose the presence of the real patient records the attacker knows.

        Parameters
        ----------
        real_data: pytrial.data.patient_sequence.PatientSequence
            A PatientSequence object that contains real patient records.
        
        syn_data: pytrial.data.patient_sequence.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.
        
        known_patient: list[int] or int
            The percentage of patient records known to the attacker.

        confidence: float
            The confidence level of the attacker. Default is 1.0, meaning the attacker decides to disclose the synthetic record
            when the synthetic record fully matches the real record (100% columns overlap). Lower confidence means the higher recall.

        prior: float
            The prior probability of the attacker. Default is 1.0, meaning all the real records owned by the attacker are
            in the training set of the generator. If the prior is less than 1.0, e.g., 0.5, it means the 50% of attacker owned records
            are in the training set of the generator, and the other 50% are not. The prior probability will affect the disclosure risk.
            Larger prior means higher disclosure risk.

        device: str
            Not used. For compatibility with other metrics.

        Usage
        -----
        >>> from pytrial.tasks.trial_simulation.sequence.evaluation import PresenceDisclosurePrivacy
        >>> score = PresenceDisclosurePrivacy.compute(real_data, syn_data)
        '''
        assert isinstance(real_data, SequencePatient), '`real_data` must be a `SequencePatient` object.'
        assert isinstance(syn_data, SequencePatient), '`syn_data` must be a `SequencePatient` object.'

        if isinstance(known_patient, int):
            known_patient = [known_patient]

        # build dataset
        real_df, syn_df = self._build_dataset(real_data, syn_data)

        # select from pids with the ration in known_patient
        pids = real_df['pid'].unique()

        if prior == 1:
            known_patient_indices = [np.random.choice(pids, size=int(len(pids)*k/100), replace=False) for k in known_patient]
        else:
            # if the prior is less than 1, we need to split the pids into two groups
            # get 1/prior of pids as the known pids
            known_patient_indices = [np.random.choice(pids, size=int(len(pids)*k/100/prior), replace=False) for k in known_patient]
            # randomly select prior ratio of pids held out from the real_df
            held_out_pids = [np.random.choice(k, size=int(prior*len(k)), replace=False) for k in known_patient_indices]

        comp_syn = syn_df.drop(['vid'], axis=1)
        scores = {}

        for i, indices in enumerate(known_patient_indices):
            comp_train = real_df[real_df['pid'].isin(indices)]
            if prior < 1:
                # should remove the held out pids from the training set
                # comp_train_tp is the number of true positive matches in the attacker dataset
                comp_train_tp = comp_train[~comp_train['pid'].isin(held_out_pids[i])]

                # find attacker records similar to the synthetic records
                matched = comp_train.isin(comp_syn).mean(1) >= confidence
                
                # compute precision: how many matched records are true positive
                prec = comp_train[matched]['pid'].isin(comp_train_tp['pid']).mean()
                scores[f'precision_known_{known_patient[i]}_conf_{confidence}_prior_{prior}'] = prec

                # compute recall: how many true positive records are matched
                rec = comp_train[matched]['pid'].isin(comp_train_tp['pid']).sum() / len(comp_train_tp)
                scores[f'recall_known_{known_patient[i]}_conf_{confidence}_prior_{prior}'] = rec

            else:
                comp_train = comp_train.drop(['vid'], axis=1)
                matched = comp_train.isin(comp_syn).mean(1) >= confidence
                match_count =  comp_train[matched]['pid'].isin(comp_syn['pid']).sum()
                scores[f'recall_known_{known_patient[i]}_conf_{confidence}_prior_{prior}'] = match_count / len(comp_train)
                scores[f'precision_known_{known_patient[i]}_conf_{confidence}_prior_{prior}'] = 1 # prior = 1, so all attacker data are in the training set, if matched, it is a true positive

        return scores

    @staticmethod
    def _build_dataset(real_data, syn_data):
        '''
        BUild a dataset for comparing the distribution of real and synthetic patient records.

        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.
        '''
        order = real_data.metadata['visit']['order']
        voc = real_data.metadata['voc']
        # build tabular format out of real and syn data
        real_df = transform_sequence_to_table(real_data.visit, order, voc)
        syn_df = transform_sequence_to_table(syn_data.visit, order, voc)
        return real_df, syn_df


class SequenceUtilityMetric(SequenceMetric):
    '''
    Base class for Machine Learning based models to learn from the synthetic patient records.
    And predict the outcome of the real patient records.
    Compared with the model learning from the real patient records.
    The output is the AUC of the prediction from the real-data-model and the synthetic-data-model.
    '''
    @staticmethod
    def _build_dataset(real_data, syn_data):
        '''
        Build a dataset for training a ML model.

        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        Returns
        -------
        train_set: torch.utils.data.Dataset
            A dataset for training a ML model.
        
        test_set: torch.utils.data.Dataset
            A dataset for testing a ML model.
        '''
        train_real_data, test_real_data = split_dataset(real_data, split=0.7)
        train_syn_data, test_syn_data = split_dataset(syn_data, split=0.7)
        return [(train_real_data, test_real_data), (train_syn_data, test_syn_data)]
    
    @staticmethod
    def _build_model(train_data, test_data):
        '''
        Build a detection model.
        '''
        raise NotImplementedError


class RNNUtilityDetection(SequenceUtilityMetric):
    @staticmethod
    def _build_model(train_data, test_data, device='cpu'):
        '''
        Build a detection model.
        '''
        model = RNN(
            vocab_size=[len(train_data.metadata['voc'][o]) for o in train_data.metadata['visit']['order']],
            orders=train_data.metadata['visit']['order'],
            mode='binary',
            bidirectional=False,
            epochs=50,
            batch_size=128,
            weight_decay=0,
            rnn_type='lstm',
            n_rnn_layer=1,
            learning_rate=5e-4,
            device=device,
            )
        model.fit(train_data=train_data, valid_data=test_data)
        return model

    @classmethod
    def compute(self, real_data, syn_data, device='cpu'):
        '''
        Compute the detection score. Report the performance of the model trained on real data (real-data-model) 
        and the model trained on synthetic data (syn-data-model).
        Performance is evaluated on the test set of the real data.

        If the syn-data-model performs equally well or better than the real-data-model, then the synthetic data has high utility.
        
        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        device: str
            The device to run the model. Default is 'cpu'.

        Returns
        -------
        float: The detection score of real-data-model and synthetic-data-model.
        '''
        assert isinstance(real_data, SequencePatient), '`real_data` must be a `SequencePatient` object.'
        assert isinstance(syn_data, SequencePatient), '`syn_data` must be a `SequencePatient` object.'

        # build a dataset with labeled real and syndata
        real_data_traintest, syn_data_traintest = self._build_dataset(real_data, syn_data)

        # train a RNN model
        real_model = self._build_model(real_data_traintest[0], real_data_traintest[1])
        syn_model = self._build_model(syn_data_traintest[0], syn_data_traintest[1], device=device)

        # evaluate the model
        outputs = {'real-data-model-auc': None, 'syn-data-model-auc': None}
        test_data = real_data_traintest[1] 

        pred = real_model.predict(test_data)
        label = test_data.get_label()
        auc = roc_auc_score(label, pred)
        outputs['real-data-model-auc'] = auc

        pred = syn_model.predict(test_data)
        label = test_data.get_label()
        auc = roc_auc_score(label, pred)
        outputs['syn-data-model-auc'] = auc

        return outputs


class SequenceFidelityMetric(SequenceMetric):
    '''
    Base class for evaluating the fidelity of synthetic patient records.
    '''
    @staticmethod
    def _build_dataset(real_data, syn_data):
        '''
        BUild a dataset for comparing the distribution of real and synthetic patient records.

        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.
        '''
        order = real_data.metadata['visit']['order']
        voc = real_data.metadata['voc']
        # build tabular format out of real and syn data
        real_df = transform_sequence_to_table(real_data.visit, order, voc)
        syn_df = transform_sequence_to_table(syn_data.visit, order, voc)
        return real_df, syn_df


class DimWiseFidelity(SequenceFidelityMetric):
    '''
    Evaluate the fidelity of synthetic patient records by computing the dimension-wise probability distributions 
    of the real and synthetic patient records. Overall results shown as the r-value.
    '''
    @classmethod
    def compute(self, real_data, syn_data, event_type=None, plot=False, save_path=None, device=None):
        '''
        Compute the fidelity score. Report the r-value of the dimension-wise probability distributions of real and synthetic patient records.

        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.
        
        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.
        
        event_type: str or list[str] or None
            The type of event to compute the fidelity score. Default is None, which means all events are considered.
            If `event_type` is not None, then the fidelity score is computed for the specified event type list.

        plot: bool
            Whether to plot the scatters distribution. Default is False.

        save_path: str
            The path to save the plot. Default is None. Only used when `plot` is True.

        device: str
            Not used. For compatibility with other metrics.

        Usage
        -----
        >>> from pytrial.tasks.trial_simulation.metrics import DimWiseFidelity
        >>> # real_data and syn_data are PatientSequence objects for all types of events
        >>> fidelity_score = DimWiseFidelity.compute(real_data, syn_data)
        >>> # real_data and syn_data are PatientSequence objects for medication events
        >>> fidelity_score = DimWiseFidelity.compute(real_data, syn_data, event_type='medication', plot=True)
        '''
        # build a dataset with labeled real and syndata
        real_df, syn_df = self._build_dataset(real_data, syn_data)
        
        if event_type is None or event_type == 'all':
            event_types = list(real_data.metadata['voc'].keys())
        else:
            event_types = event_type if isinstance(event_type, list) else [event_type]

        outputs = {}
        syn_prob_outputs = {}
        real_prob_outputs = {}
        for etype in event_types:
            # compute the real marginal probability distribution
            real_probs = self._compute_marginal_probability(real_df, event_type=etype)
            # compute the synthetic marginal probability distribution
            syn_probs = self._compute_marginal_probability(syn_df, event_type=etype)
            # compute the r-value
            r_value = np.corrcoef(real_probs, syn_probs)[0, 1]
            # save the r-value
            outputs[etype] = r_value
            # save the real and synthetic probability distributions
            syn_prob_outputs[etype] = syn_probs
            real_prob_outputs[etype] = real_probs

        # plot the distribution
        if plot:
            for etype in event_types:
                r_value = outputs[etype]
                real_probs = real_prob_outputs[etype]
                syn_probs = syn_prob_outputs[etype]
                self._plot_scatter(real_probs, syn_probs, r_value)
                if save_path is not None:
                    plt.savefig(save_path + f'/{etype}.png')
                plt.show()
        return outputs

    @staticmethod
    def _compute_marginal_probability(df, event_type=None):
        '''
        Compute the marginal probability distribution of a dataframe.
        '''
        binary_probability = []
        if event_type is not None:
            columns = [c for c in df.columns if event_type in c]
        else:
            columns = df.columns
        for c in columns:
            binary_probability.append(df[c].mean())
        return binary_probability
    
    @staticmethod
    def _plot_scatter(real_probs, syn_probs, r_value):
        '''
        Plot the scatter distribution of real and synthetic data.
        '''
        import matplotlib.pyplot as plt
        import matplotlib
        plt.rcParams["figure.figsize"] = (3,3)
        SMALL_SIZE = 14
        matplotlib.rc('font', size=SMALL_SIZE)
        matplotlib.rc('axes', titlesize=SMALL_SIZE)
        plt.scatter(real_probs, syn_probs, label=f'r = {np.round(r_value, 2)}')
        plt.legend(loc=1)


from collections import defaultdict
import warnings
import pdb, os
import json
from typing import Any, Dict, List, Optional, Tuple, Union

from dataclasses import dataclass
import torch
import numpy as np
from transformers import BartTokenizer
from transformers import BartConfig
from transformers.file_utils import ModelOutput
from tokenizers import Tokenizer
from tokenizers.pre_tokenizers import Whitespace
from tokenizers.models import WordLevel

from . import constants

def EHRBartConfig(data_tokenizer, model_tokenizer, **kwargs):
    '''Build the config used for building the promptBart model.
    '''
    bart_config = BartConfig.from_pretrained('facebook/bart-base')
    kwargs.update(model_tokenizer.get_num_tokens)
    kwargs['data_tokenizer_num_vocab'] = len(data_tokenizer)
    if 'd_prompt_hidden' not in kwargs:
        kwargs['d_prompt_hidden'] = 128
    if 'n_num_feature' not in kwargs:
        kwargs['n_num_feature'] = None
    if 'cat_cardinalities' not in kwargs:
        kwargs['cat_cardinalities'] = None
    bart_config.__dict__.update(kwargs)

    # specify bos, eos token id
    bart_config.__dict__['decoder_start_token_id'] = 0
    bart_config.__dict__['bos_token_id'] = 0
    bart_config.__dict__['eos_token_id'] = 1
    bart_config.__dict__['forced_eos_token_id'] = 1
    return bart_config

class DataTokenizer(BartTokenizer):
    r'''construct tokenizer to process the input raw records.
    '''
    new_token_type_list = constants.CODE_TYPES
    special_token_dict = constants.SPECIAL_TOKEN_DICT
    code_vocab = defaultdict(list)

    def add_token_to_code_vocab(self, tokens, code):
        self.add_tokens(tokens)

        if code not in self.code_vocab:
            self.code_vocab[code] = np.array(tokens)
        else:
            origin_tokens = self.code_vocab[code]
            new_tokens = np.array(tokens)
            self.code_vocab[code] = np.unique(np.concatenate([origin_tokens, new_tokens]))

    def update_special_token_config(self, code_types):
        self.new_token_type_list = code_types
        self.special_token_dict = {}
        special_token_list = []
        for code_type in code_types:
            l = [f'<{code_type}>', f'</{code_type}>']
            self.special_token_dict[code_type] = l
            special_token_list.extend(l)
        self.add_tokens(special_token_list)

    def extend_vocab(self, token_dict):
        '''
        Parameters:
        ----------
        token_dict: dict
            key: code type, value: a list of tokens.
        '''
        for key in token_dict.keys():
            self.code_vocab[key] = np.array(token_dict[key])
            self.add_tokens(token_dict[key])

    def extend_vocab_from_dir(self, data_dir):
        # add new tokens from the data dir
        for key in self.new_token_type_list:
            filename = os.path.join(data_dir,'{}_token_list.txt'.format(key))
            with open(filename, 'r', encoding='utf-8') as f:
                token_list = [line.strip() for line in f.readlines()]
            self.code_vocab[key] = np.array(token_list)
            self.add_tokens(token_list)

        # add special tokens indicating different modality
        for key, value in self.special_token_dict.items():
            self.add_tokens(value, special_tokens=True)

class ModelTokenizer:
    r'''construct an EHR tokenizer that converts tokenized indices to code-specific token indices.
    '''
    def __init__(self, tokenizer: DataTokenizer):
        # map_token = lambda x: str(tokenizer(x).input_ids[1])
        org_vocab = tokenizer.get_vocab()
        tokenizer_dict = {}
        num_token_dict = {}
        for key, value in tokenizer.code_vocab.items():
            vocab = defaultdict(int)
            vocab[constants.UNKNOWN_TOKEN] = 0
            for i,token in enumerate(tokenizer.special_token_dict[key]):
                vocab[str(org_vocab[token])] = i+1
            offset = len(vocab)

            for i, token in enumerate(value): # str token = 'diag_xxx'
                # fix: if token has more than one '_', e.g., 'diag_t_a_b_100', will only take the last '100' as the index. 
                # _, index = token.split('_')
                indexes = token.split('_')
                try:
                    index = int(indexes[-1])
                except:
                    raise ValueError(f"Token {token} is not a valid token, it should be splited by '_' and the last part should be a number, e.g., 'diag_100'. ")
                vocab[str(org_vocab[token])] = index + offset
            
            # new tokenizer
            specific_tokenizer = Tokenizer(WordLevel(vocab=vocab, unk_token=constants.UNKNOWN_TOKEN))
            specific_tokenizer.pre_tokenizer = Whitespace()

            # num_token_dict is decided by the max index instead of number of tokens
            num_token_dict[key] = (max(vocab.values())+1) - offset
            tokenizer_dict[key] = specific_tokenizer

        # each code type has its own tokenizer corresponding to specific LM heads
        self.tokenizer_dict = tokenizer_dict
        self.num_token_dict = num_token_dict
        self.label_offset = offset


    def encode(self, input_ids, code_type):
        if len(input_ids.shape) > 1: # a batch
            ids = self.encode_batch(input_ids, code_type)
        else:
            ids = self.tokenizer_dict[code_type].encode(input_ids.cpu().numpy().astype(str), is_pretokenized=True).ids
            ids = torch.tensor(ids, device=input_ids.device)
        return ids

    def encode_batch(self, input_ids, code_type):
        ids_list = self.tokenizer_dict[code_type].encode_batch(input_ids.cpu().numpy().astype(str).tolist(), is_pretokenized=True)

        ids = torch.tensor([x.ids for x in ids_list], device=input_ids.device)
        return ids

    @property
    def get_num_tokens(self):
        return self.num_token_dict

@dataclass
class EHRBartOutput(ModelOutput):
    """
    Base class for sequence-to-sequence language models outputs.

    Args:
        loss (:obj:`torch.FloatTensor` of shape :obj:`(1,)`, `optional`, returned when :obj:`labels` is provided):
            Language modeling loss.
        logits (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, config.vocab_size)`):
            Prediction scores of the language modeling head (scores for each vocabulary token before SoftMax).
        past_key_values (:obj:`tuple(tuple(torch.FloatTensor))`, `optional`, returned when ``use_cache=True`` is passed or when ``config.use_cache=True``):
            Tuple of :obj:`tuple(torch.FloatTensor)` of length :obj:`config.n_layers`, with each tuple having 2 tensors
            of shape :obj:`(batch_size, num_heads, sequence_length, embed_size_per_head)`) and 2 additional tensors of
            shape :obj:`(batch_size, num_heads, encoder_sequence_length, embed_size_per_head)`.

            Contains pre-computed hidden-states (key and values in the self-attention blocks and in the cross-attention
            blocks) that can be used (see :obj:`past_key_values` input) to speed up sequential decoding.
        decoder_hidden_states (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_hidden_states=True`` is passed or when ``config.output_hidden_states=True``):
            Tuple of :obj:`torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer)
            of shape :obj:`(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the decoder at the output of each layer plus the initial embedding outputs.
        decoder_attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_attentions=True`` is passed or when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape :obj:`(batch_size, num_heads,
            sequence_length, sequence_length)`.

            Attentions weights of the decoder, after the attention softmax, used to compute the weighted average in the
            self-attention heads.
        cross_attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_attentions=True`` is passed or when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape :obj:`(batch_size, num_heads,
            sequence_length, sequence_length)`.

            Attentions weights of the decoder's cross-attention layer, after the attention softmax, used to compute the
            weighted average in the cross-attention heads.
        encoder_last_hidden_state (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, hidden_size)`, `optional`):
            Sequence of hidden-states at the output of the last layer of the encoder of the model.
        encoder_hidden_states (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_hidden_states=True`` is passed or when ``config.output_hidden_states=True``):
            Tuple of :obj:`torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer)
            of shape :obj:`(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the encoder at the output of each layer plus the initial embedding outputs.
        encoder_attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_attentions=True`` is passed or when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape :obj:`(batch_size, num_heads,
            sequence_length, sequence_length)`.

            Attentions weights of the encoder, after the attention softmax, used to compute the weighted average in the
            self-attention heads.
        perplexity:
            perplexity calculated when the label mask is given.
    """

    loss: Optional[torch.FloatTensor] = None
    logits: torch.FloatTensor = None
    past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    decoder_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    decoder_attentions: Optional[Tuple[torch.FloatTensor]] = None
    cross_attentions: Optional[Tuple[torch.FloatTensor]] = None
    encoder_last_hidden_state: Optional[torch.FloatTensor] = None
    encoder_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    encoder_attentions: Optional[Tuple[torch.FloatTensor]] = None
    perplexity: Optional[torch.FloatTensor] = None



import pdb

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.nn import LSTMCell
from torch.autograd import Variable
from torch.nn.utils.rnn import pad_sequence, pack_padded_sequence, pad_packed_sequence
import numpy as np

from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceIndivBase
from ..trainer import IndivSeqTrainer

class RaimExtract(nn.Module):

    def __init__(self, input_size, window_size, hidden_size):
        super(RaimExtract, self).__init__()

        self.input_size = input_size
        self.window_size = window_size
        self.hidden_size = hidden_size

        self.w_h_alpha = nn.Linear(self.hidden_size, self.window_size)
        self.w_a_alpha = nn.Linear(self.input_size, 1)

        self.w_h_beta = nn.Linear(self.hidden_size, self.input_size)
        self.w_a_beta = nn.Linear(self.window_size, 1)

        self.activate_func = nn.Tanh()
        self.weights_func = nn.Softmax(dim = -1)

    def forward(self, input_data, h_t_1):
        # shape of input_data : <n_batch, window_size, input_size> 
        # shape of h_t_1      : <n_batch, hidden_size>
        # shape of _alpha_weights : <n_batch, window_size> 
        _alpha_weights = self.activate_func(self.w_h_alpha(h_t_1) +\
                                            self.w_a_alpha(input_data.reshape(-1, self.input_size)).\
                                            squeeze().\
                                            reshape(-1, self.window_size)
                                           )
        _alpha_weights = self.weights_func(_alpha_weights)

        # shape of _beta_weights : <n_batch, input_size> 
        _beta_weights = self.activate_func(self.w_h_beta(h_t_1) +\
                                           self.w_a_beta(input_data.permute(0, 2, 1).\
                                                         reshape(-1, self.window_size)).\
                                           squeeze().\
                                           reshape(-1, self.input_size)
                                          )
        _beta_weights = self.weights_func(_beta_weights)

        # shape of _alpha_weights_v : <n_batch, window_size, 1> 
        _alpha_weights_v = _alpha_weights.unsqueeze(-1)
        # shape of _beta_weights_v  : <n_batch, 1, input_size> 
        _beta_weights_v = _beta_weights.unsqueeze(1)

        # shape of weight_A  : <n_batch, window_size, input_size> 
        weight_A = _alpha_weights_v * _beta_weights_v

        # shape of output_v  : <n_batch, input_size> 
        output_v = torch.sum(weight_A * input_data, dim = 1)
        
        return output_v

class BuildModel(nn.Module):
    def __init__(self, 
                 input_size = None,
                 window_size = 3,
                 hidden_size = 16,
                 output_size = 8,
                 label_size = 1,
                 device = None):
        super(BuildModel, self).__init__()
        assert input_size != None and isinstance(input_size, int), 'fill in correct input_size' 
        self.input_size = input_size
        self.window_size = window_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.label_size = label_size
        self.rnn_raimf = RaimExtract(input_size, window_size, hidden_size)
        self.rnn_unit = LSTMCell(input_size, hidden_size) 
        self.predict_func = nn.Linear(self.output_size, self.label_size)
        self.pad = nn.ConstantPad1d((self.window_size-1, 0), 0.)
        self.device = device

    def forward(self, input_data):
        """
        
        Parameters
        
        ----------
        input_data = {
                      'X': shape (batchsize, n_timestep, n_featdim)
                      'M': shape (batchsize, n_timestep), the first visit to the current visit are all 1, the later are 0, i.e., [1,1,1,0,0].
                      'cur_M': shape (batchsize, n_timestep), only the current visit is 1 others are 0, i.e.,[0,0,1,0,0]
                      'T': shape (batchsize, n_timestep), the time interval between visits, i.e., for timestamp [0, 1, 2, 3, 4], T = [0, 1, 1, 1, 1].
                     }
        
        Return
        
        ----------
        
        all_output, shape (batchsize, n_timestep, n_labels)
            
            predict output of each time step
            
        cur_output, shape (batchsize, n_labels)
        
            predict output of last time step
        
        """
        X = input_data['v']
        
        if 'v_lengths' in input_data:
            v_len = input_data['v_lengths']
            mask, cur_mask = self._create_mask_from_visit_length(v_len=v_len)
            mask = mask.to(X.device)
            cur_mask = cur_mask.to(X.device)
        else:
            mask = cur_mask = None

        batchsize, n_timestep, n_f = X .shape
        # shape of X     : <batchsize, n_timestep, n_featdim>        
        # shape of pad_X : <batchsize, n_timestep + window_size - 1, n_featdim>        
        pad_X = self.pad(X.permute(0,2,1)).permute(0,2,1)
        
        h0 = Variable(torch.zeros(batchsize, self.hidden_size)).to(self.device)
        c0 = Variable(torch.zeros(batchsize, self.hidden_size)).to(self.device)
        outs = []        
        hn = h0
        cn = c0
        for i in range(self.window_size-1, n_timestep+self.window_size-1):
            cur_x = pad_X[ : , i - self.window_size + 1 : i + 1, : ]
            z_value = self.rnn_raimf(cur_x, hn)
            hn, cn = self.rnn_unit(z_value, (hn, cn))
            outs.append(hn)
        outputs = torch.stack(outs, dim =1)
        n_batchsize, n_timestep, n_featdim = outputs.shape
        all_output =  self.predict_func(outputs.reshape(n_batchsize*n_timestep, n_featdim)).\
                    reshape(n_batchsize, n_timestep, self.label_size) # bs, timestep, 1

        if mask is not None and cur_mask is not None:
            all_output *= mask.unsqueeze(-1)
            cur_output = (all_output * cur_mask.unsqueeze(-1)).sum(dim=1)
        else:
            cur_output = all_output.sum(1) 

        # TODO: all_output is the per visit label prediction, currently we only support patient-level prediction.
        return cur_output

    def _create_mask_from_visit_length(self, v_len):
        mask = torch.zeros(len(v_len),max(v_len))
        cur_mask = torch.zeros(len(v_len),max(v_len))

        for i in range(len(v_len)):
            mask[i, :v_len[i]] = 1
            cur_mask[i, v_len[i]-1] = 1
        return mask, cur_mask

class RAIM(SequenceIndivBase):
    '''
    Implment RAIM for longitudinal patient records predictive modeling [1]_.

    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.
    
    orders: list[str]
        A list of orders when treating inputs events. Should have the same shape of `vocab_size`.

    mode: str
        Prediction traget in ['binary','multiclass','multilabel','regression'].

    output_dim: int
        Output dimension of the model.

        - If binary classification, output_dim=1;
        - If multiclass/multilabel classification, output_dim=n_class;
        - If regression, output_dim=1.

    max_visit: int
        The maximum number of visits for input event codes.

    window_size: int
        The window size in the recurrent part of RAIM.
    
    hidden_size: int
        Embedding size for encoding input event codes.
    
    hidden_output_size: int
        The output size of the intermediate layers. Not the output_dim.

    dropout: float
        Dropout rate in the intermediate layers.
    
    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.
    
    experiment_id: str
        The prefix when saving the checkpoints during the training.
    
    Notes
    -----
    .. [1] Xu, Y., Biswal, S., Deshpande, S. R., Maher, K. O., & Sun, J. (2018, July). 
        Raim: Recurrent attentive and intensive model of multimodal patient monitoring data. 
        In Proceedings of the 24th ACM SIGKDD international conference on Knowledge Discovery & Data Mining (pp. 2565-2573).

    '''
    def __init__(self, 
        vocab_size,
        orders,
        mode=None, 
        output_dim=None,
        max_visit=None,
        window_size=3,
        hidden_size=8,
        hidden_output_size=8,
        dropout=0.5,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test', 
        ):
        super().__init__(experiment_id, mode, output_dim)
        self.config = {
            'mode':self.mode,
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'window_size':window_size,
            'hidden_size':hidden_size,
            'hidden_output_size':hidden_output_size,
            'output_dim':self.output_dim,
            'dropout':dropout,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':orders,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()

    def load_model(self, checkpoint):
        '''
        Load pretrained model from the disk.

        Parameters
        ----------
        checkpoint: str
            The input directory that stores the trained pytorch model and configuration.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        
        self.model.load_state_dict(state_dict['model'])

    def save_model(self, output_dir):
        '''
        Save the pretrained model to the disk.

        Parameters
        ----------
        output_dir: str
            The output directory that stores the trained pytorch model and configuration.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model.state_dict()}, output_dir=output_dir)

    def fit(self, train_data, valid_data=None):
        '''
        Train model with sequential patient records.
        
        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events; 'y' corresponds to labels.

        valid_data: SequencePatientBase
            A `SequencePatientBase` contains patient records used to make early stopping of the
            model.
        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data, valid_data=valid_data)


    def predict(self, test_data):
        '''
        Predict patient outcomes using longitudinal trial patient sequences.
        
        Parameters
        ----------
        test_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        '''
        test_dataloader = self.get_test_dataloader(test_data)
        outputs = self._predict_on_dataloader(test_dataloader)
        return outputs['pred']

    def _build_model(self):
        self.model = BuildModel(
            input_size=self.config['total_vocab_size'],
            window_size=self.config['window_size'],
            hidden_size=self.config['hidden_size'],
            output_size=self.config['hidden_output_size'],
            label_size=self.config['output_dim'],
            device=self.config['device'],
            )
        self.model.to(self.device)

    def _fit_model(self, train_data, valid_data=None):
        test_metric_dict = {
            'binary': 'auc',
            'multiclass': 'acc',
            'regression': 'mse',
            'multilabel': 'f1', # take average of F1 scores
            }
        train_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()

        if self.config['mode'] == 'regression':
            less_is_better = True
        else:
            less_is_better = False
    
        train_objectives = [(train_dataloader, loss_model) for loss_model in loss_models]
        trainer = IndivSeqTrainer(model=self,
            train_objectives=train_objectives,
            test_data=valid_data,
            test_metric=test_metric_dict[self.config['mode']],
            less_is_better=less_is_better,
            )
        trainer.train(**self.config)